//
//  RootViewController.m
//  Locations
//
//  Created by Chao-Hsin Shih on 9/22/10.
//  Copyright 2010 sjsu. All rights reserved.
//

#import "RootViewController.h"
#import "Event.h"
#import "Photo.h"
#import "CameraDemoViewController.h"
#import "EXF.h"

@implementation RootViewController
@synthesize eventsArray;
@synthesize managedObjectContext;
@synthesize addButton;
@synthesize locationManager,compass ;
@synthesize picker;
@synthesize coordination, compassXY, compassZ;
#pragma mark -
#pragma mark View lifecycle


- (void)viewDidLoad {
	[super viewDidLoad];
	
    // Set the title.
    self.title = @"Locations";
	
    // Set up the buttons.
    self.navigationItem.leftBarButtonItem = self.editButtonItem;
	
    addButton = [[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd  target:self action:@selector(addEvent)];
    addButton.enabled = NO;
    self.navigationItem.rightBarButtonItem = addButton;
	
    // Start the location manager.
    [[self locationManager] startUpdatingLocation];
	
	//initializa the eventsArray to store the data
	//eventsArray = [[NSMutableArray alloc] init];
	
	 //create a fetch request and set the entity
	 NSFetchRequest *request = [[NSFetchRequest alloc] init];
	 NSEntityDescription *entity = [NSEntityDescription entityForName:@"Event" inManagedObjectContext:managedObjectContext];
	[request setEntity:entity];
	 
	 // Order the events by creation date, most recent first.
	 NSSortDescriptor *sortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"creationDate" ascending:NO];
	 NSArray *sortDescriptors = [[NSArray alloc] initWithObjects:sortDescriptor, nil];
	 [request setSortDescriptors:sortDescriptors];
	 [sortDescriptor release];
	 [sortDescriptors release];
	 
	 // Execute the fetch -- create a mutable copy of the result.
	 NSError *error = nil;
	 NSMutableArray *mutableFetchResults = [[managedObjectContext executeFetchRequest:request error:&error] mutableCopy];
	 if (mutableFetchResults == nil) {
	 // Handle the error.
	 }
	 
	 // Set self's events array to the mutable array, then clean up.
	 [self setEventsArray:mutableFetchResults];
	 [mutableFetchResults release];
	 [request release];

}

#pragma mark -
#pragma mark read GPS and Accelerometer
- (CLLocationManager *)locationManager {
	
    if (locationManager != nil) {
        return locationManager;
    }
	
    locationManager = [[CLLocationManager alloc] init];
    locationManager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters;
    locationManager.delegate = self;
	
    return locationManager;
}

- (void)locationManager:(CLLocationManager *)manager
    didUpdateToLocation:(CLLocation *)newLocation
           fromLocation:(CLLocation *)oldLocation {
    addButton.enabled = YES;
	NSString *coorString = [[NSString alloc] initWithFormat:@"Latitude: %g° , Longitude: %g°", newLocation.coordinate.latitude,newLocation.coordinate.longitude];
	coordination.text = coorString;
	[coorString release];
}

- (void)locationManager:(CLLocationManager *)manager
       didFailWithError:(NSError *)error {
	
	NSString *errorType = (error.code == kCLErrorDenied) ? @"Access Denied" : @"Unknown Error";
	UIAlertView *alert = [[UIAlertView alloc] initWithTitle:@"Error getting Location"
													message:errorType delegate:nil   cancelButtonTitle:@"Okay" otherButtonTitles:nil]; 
	[alert show]; 
	[alert release];
	
    addButton.enabled = NO;
}

#pragma mark - 
- (void)accelerometer:(UIAccelerometer *)accelerometer 
        didAccelerate:(UIAcceleration *)acceleration {
    
	
	NSString *acc = [[NSString alloc] initWithFormat:@"z: %g° ", acceleration.z];
	compassZ.text = acc; 
	[acc release];
	
	NSString *accxy = [[NSString alloc] initWithFormat:@"acc: x: %g y: %g° ", acceleration.x, acceleration.y];
	compassXY.text = acc; 
	[accxy release];
	self.compass = acceleration;
}

#pragma mark -
#pragma mark Initialization

/*
- (id)initWithStyle:(UITableViewStyle)style {
    // Override initWithStyle: if you create the controller programmatically and want to perform customization that is not appropriate for viewDidLoad.
    if ((self = [super initWithStyle:style])) {
    }
    return self;
}
*/



- (void)viewWillAppear:(BOOL)animated {
    [super viewWillAppear:animated];
	[self.tableView reloadData];

}

/*
- (void)viewDidAppear:(BOOL)animated {
    [super viewDidAppear:animated];
}
*/
/*
- (void)viewWillDisappear:(BOOL)animated {
    [super viewWillDisappear:animated];
}
*/
/*
- (void)viewDidDisappear:(BOOL)animated {
    [super viewDidDisappear:animated];
}
*/

// Override to allow orientations other than the default portrait orientation.
- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation {

	return YES;
}



#pragma mark -
#pragma mark Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {
    // Return the number of sections.
    return 1;
}


- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section {
    // Return the number of rows in the section.
    return [eventsArray count];
}


// Customize the appearance of table view cells.
- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath {
	//set the style for displaying info in table
    static NSDateFormatter *dateFormatter = nil;
	if (dateFormatter == nil) {
		dateFormatter = [[NSDateFormatter alloc]init];
		[dateFormatter setTimeStyle:NSDateFormatterMediumStyle];
		[dateFormatter setDateStyle:NSDateFormatterMediumStyle];
	}
	
	static NSNumberFormatter *numberFormatter = nil;
	if(numberFormatter == nil)
	{
		numberFormatter = [[NSNumberFormatter alloc]init];
		[numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
		[numberFormatter setMaximumFractionDigits:3];
	}
	
	static NSString *cellIdentifier = nil;
	
	 
	UITableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:cellIdentifier];
	if(cell == nil){
		cell = [[[UITableViewCell alloc]initWithStyle:UITableViewCellStyleSubtitle reuseIdentifier:cellIdentifier] autorelease];
		cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
	}
	
	
	//set the table cell by pulling data from Event database
	Event *event = (Event *) [eventsArray objectAtIndex:indexPath.row];
	
	cell.textLabel.text = [dateFormatter stringFromDate:[event creationDate]];
	cell.imageView.image = event.thumbnail;
	
	NSString *string = [NSString stringWithFormat:@"%@, %@", [numberFormatter stringFromNumber:[event latitude]], [numberFormatter stringFromNumber:[event longitude]]];
	
	cell.detailTextLabel.text = string;
	return cell;
}


/*
// Override to support conditional editing of the table view.
- (BOOL)tableView:(UITableView *)tableView canEditRowAtIndexPath:(NSIndexPath *)indexPath {
    // Return NO if you do not want the specified item to be editable.
    return YES;
}
*/



// Override to support editing the table view.
- (void)tableView:(UITableView *)tableView commitEditingStyle:(UITableViewCellEditingStyle)editingStyle forRowAtIndexPath:(NSIndexPath *)indexPath {
    
    if (editingStyle == UITableViewCellEditingStyleDelete) {
		
		//Delete the managed object at the given index path
		NSManagedObject *eventtToDelete = [eventsArray objectAtIndex:indexPath.row];
		[managedObjectContext deleteObject:eventtToDelete];
		
		//update array and table view
		[eventsArray removeObjectAtIndex:indexPath.row];
		[tableView deleteRowsAtIndexPaths:[NSArray arrayWithObject:indexPath] withRowAnimation:YES];
		
		//commit the change
		NSError *error;
		if(![managedObjectContext save:&error]){
			NSLog(@"row delete failed");	
		}
    }   
       
}



/*
// Override to support rearranging the table view.
- (void)tableView:(UITableView *)tableView moveRowAtIndexPath:(NSIndexPath *)fromIndexPath toIndexPath:(NSIndexPath *)toIndexPath {
}
*/


/*
// Override to support conditional rearranging of the table view.
- (BOOL)tableView:(UITableView *)tableView canMoveRowAtIndexPath:(NSIndexPath *)indexPath {
    // Return NO if you do not want the item to be re-orderable.
    return YES;
}
*/

/*
#pragma mark -
#pragma mark Table view delegate

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath {
    // Navigation logic may go here. Create and push another view controller.
	
	 <#DetailViewController#> *detailViewController = [[<#DetailViewController#> alloc] initWithNibName:@"<#Nib name#>" bundle:nil];
     // ...
     // Pass the selected object to the new view controller.
	 [self.navigationController pushViewController:detailViewController animated:YES];
	 [detailViewController release];
	 
}
*/

#pragma mark -
#pragma mark geotagImage using EXF type

// Helper methods for location conversion 
-(NSMutableArray*) createLocArray:(double) val{ 
	val = fabs(val); 
	NSMutableArray* array = [[NSMutableArray alloc] init]; 
	double deg = (int)val; 
	[array addObject:[NSNumber numberWithDouble:deg]]; 
	val = val - deg; 
	val = val*60; 
	double minutes = (int) val; 
	[array addObject:[NSNumber numberWithDouble:minutes]]; 
	val = val - minutes; 
	val = val*60; 
	double seconds = val; 
	[array addObject:[NSNumber numberWithDouble:seconds]]; 
	return array; 
} 


-(void) populateGPS:(EXFGPSLoc *) gpsLoc :(NSArray *)locArray{ 
	long numDenumArray[2]; 
	long* arrPtr = numDenumArray; 
	[EXFUtils convertRationalToFraction:&arrPtr :[locArray objectAtIndex:0]]; 
	EXFraction* fract = [[EXFraction alloc] initWith:numDenumArray[0]:numDenumArray[1]]; 
	gpsLoc.degrees = fract; 
	[fract release]; 
	[EXFUtils convertRationalToFraction:&arrPtr :[locArray objectAtIndex:1]]; 
	fract = [[EXFraction alloc] initWith:numDenumArray[0] :numDenumArray[1]]; 
	gpsLoc.minutes = fract; 
	[fract release]; 
	[EXFUtils convertRationalToFraction:&arrPtr :[locArray objectAtIndex:2]]; 
	fract = [[EXFraction alloc] initWith:numDenumArray[0] :numDenumArray[1]]; 
	gpsLoc.seconds = fract; 
	[fract release];
}

-(NSData*) geotagImage:(UIImage*)image withLocation:(CLLocation*)imageLocation {
	NSData* jpegData =  UIImageJPEGRepresentation(image, 0.8);
	EXFJpeg* jpegScanner = [[EXFJpeg alloc] init];
	[jpegScanner scanImageData: jpegData];
	EXFMetaData* exifMetaData = jpegScanner.exifMetaData;
	// end of helper methods 
	// adding GPS data to the Exif object 
	NSMutableArray* locArray = [self createLocArray:imageLocation.coordinate.latitude]; 
	EXFGPSLoc* gpsLoc = [[EXFGPSLoc alloc] init]; 
	[self populateGPS: gpsLoc :locArray]; 
	[exifMetaData addTagValue:gpsLoc forKey:[NSNumber numberWithInt:EXIF_GPSLatitude] ]; 
	[gpsLoc release]; 
	[locArray release]; 
	locArray = [self createLocArray:imageLocation.coordinate.longitude]; 
	gpsLoc = [[EXFGPSLoc alloc] init]; 
	[self populateGPS: gpsLoc :locArray]; 
	[exifMetaData addTagValue:gpsLoc forKey:[NSNumber numberWithInt:EXIF_GPSLongitude] ]; 
	[gpsLoc release]; 
	[locArray release];
	NSString* ref;
	if (imageLocation.coordinate.latitude <0.0)
		ref = @"S"; 
	else
		ref =@"N"; 
	[exifMetaData addTagValue: ref forKey:[NSNumber numberWithInt:EXIF_GPSLatitudeRef] ]; 
	if (imageLocation.coordinate.longitude <0.0)
		ref = @"W"; 
	else
		ref =@"E"; 
	[exifMetaData addTagValue: ref forKey:[NSNumber numberWithInt:EXIF_GPSLongitudeRef] ]; 
	NSMutableData* taggedJpegData = [[NSMutableData alloc] init];
	[jpegScanner populateImageData:taggedJpegData];
	[jpegScanner release];
	//NSLog(@"exif info: %@", taggedJpegData);
	return [taggedJpegData autorelease];
}







#pragma mark -
#pragma mark camera action



- (void)addEvent
{
		if(picker == nil){
		picker = [[UIImagePickerController alloc] init];
		picker.delegate = self;
		picker.sourceType = UIImagePickerControllerSourceTypeCamera;
	}
	[self presentModalViewController:picker animated:YES];
	

	//[self showImagePicker:UIImagePickerControllerSourceTypeCamera];
	
}

-(void) imagePickerController:(UIImagePickerController *)picker 
didFinishPickingMediaWithInfo:(NSDictionary *)info{
		
		for(id key in info){
			NSLog(@"key: %@", key);
			NSLog(@"Value: %@", [info valueForKey:key]);
			UIImage *image = [info valueForKey:key];
			if([key isEqualToString: @"UIImagePickerControllerOriginalImage"]){
			
				CLLocation *location = [locationManager location];
				if(!location)
					return;

				
				Event *event = (Event *)[NSEntityDescription insertNewObjectForEntityForName:@"Event" inManagedObjectContext:managedObjectContext];
				
				CLLocationCoordinate2D coordinate = [location coordinate];
				[event setLatitude:[NSNumber numberWithDouble:coordinate.latitude]];
				[event setLongitude:[NSNumber numberWithDouble:coordinate.longitude]];
				[event setCreationDate:[NSDate date]];
				
				[event setAsisX:[NSNumber numberWithDouble:compass.x]];
				[event setAsisY:[NSNumber numberWithDouble:compass.y]];
				[event setAsisZ:[NSNumber numberWithDouble:compass.z]];
				//[acceleration release];
				//delete picture if the image already existed	
				if (event.photo) {
					[managedObjectContext deleteObject:event.photo];
				}
				//create a new photo object and set the image				
				Photo *photo = [NSEntityDescription insertNewObjectForEntityForName:@"Photo" inManagedObjectContext:managedObjectContext];
				[photo setImage:image];
				[photo setImageWithEXF:[self geotagImage:image withLocation:location]];
				
				
				//Associate the photo object with event
				[event setPhoto:photo];
				
				//Create thumbnail version of the image for the event object.
				CGSize size = image.size;
				CGFloat ratio = 0; 
				if (size.width > size.height) {
					ratio = 44.0 / size.width;
				}
				else {
					ratio = 44.0 / size.height;
				}
				
				CGRect rect = CGRectMake(0.0, 0.0, ratio * size.width, ratio * size.height);
				
				UIGraphicsBeginImageContext(rect.size);
				[image drawInRect:rect];
				event.thumbnail = UIGraphicsGetImageFromCurrentImageContext();
				UIGraphicsEndImageContext();
				
				//commit change
				NSError *error;
				
				if (![managedObjectContext save:&error]) {
					NSLog(@"got error");
				}
				//update tableviw
				[eventsArray insertObject:event  atIndex:0];
				
				NSIndexPath *indexPath = [NSIndexPath indexPathForRow:0 inSection:0];
				[self.tableView insertRowsAtIndexPaths:[NSArray arrayWithObject:indexPath] withRowAnimation:UITableViewRowAnimationFade];
				[self.tableView scrollToRowAtIndexPath:[NSIndexPath indexPathForRow:0 inSection:0] atScrollPosition:UITableViewScrollPositionTop animated:YES];
				
				[self.tableView reloadData];
				
				
		}
		//remove cameraPickerController view from the top of the screen
		[self dismissModalViewControllerAnimated:YES];
	}
}


										
										
										
#pragma mark -
#pragma mark Memory management

- (void)didReceiveMemoryWarning {
    // Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
    
    // Relinquish ownership any cached data, images, etc that aren't in use.
}
- (void)viewDidUnload {
    self.eventsArray = nil;
    self.locationManager = nil;
    self.addButton = nil;
	self.coordination = nil;
	self.compassXY = nil;
	self.compassZ = nil;
	self.compass = nil;
}

- (void)dealloc {
    [managedObjectContext release];
    [eventsArray release];
	[coordination release];
	[compassZ release];
	[compassXY release];
    [locationManager release];
    [addButton release];
	[compass release];
	
    [super dealloc];
}



@end

